/* ********************************************************************
 * Copyright (c) 2022 Biological Modeling and Visualization Laboratory,
 * University of Calgary. All rights reserved.
 * ********************************************************************/

#include <cmath>
#include <algorithm>
#include <lpfgall.h>

#define DISPLAY_STEP 1
// use Euclidean distance in Phase 3 as a speed up
#define EUCLIDEAN_PHASE_3 1 // 0 - off, 1 - on

// constants
const int MAX_PRIMORDIA_TEST = 100; // maximum number of primordia to check in Hofmeister test

const double RadToDeg = 57.29577951;
const double TwoPi = 6.283185307;
const double PHI = 0.61803398875;

const double k1 = 0.02266; // lateral displacement parameter for young primordia

const double OFFSET_RATE = 0.001*1.05;

const double DAGE = 1.0; // age increment per time step

const double tNorm = 1600.0; // normalizing time factor in calls to func()

// minimum threshold distance for insertion
double thresholdDistance;
const double initialThresholdDistance = 3.5*M_PI;
#define USER_FUNC_TH_DIST(t) func(thDistNormalized,t/(tNorm))

// active ring
const double DS = 0.0001; // arc-length step for scanning intervals
const double maxRingRadius = 1.1*1.45; // gives active ring length of 32.8327

#define USER_FUNC_AR_RADIUS(t) func(activeRingPos,t/(tNorm)) * maxRingRadius

// simulation time
const double dProgress = 1.;   // time step increment per simulation step (dt)
const double EPS = 1e-5;
double Progress = 0.;               // progress of development (from 0 to 1)

const int NUM_BRACTS = 34; // number of bract primordia in this model


// declaration of a primordium 

struct primordium{
	double rho;	// primordium radius
    double s;   // arc-length coordinate on generating curve
    double n;   // offset in the normal direction from the generating curve
    V3d normal; // normal direction from generating curve
	V3d pos;	// Cartesian position
	double age;
	int neighbour_left;
	int neighbour_right;
	double inserted_dist; // threshold distance at time of insertion
};

primordium curve_P[2000]; // array of primordia centers formed so far
primordium curve_tp;      // trial/incipient primordium
int curve_n; // number of primordia formed so far;

// curve functions

static double switch_time = 0.; // saves time when receptacle starts shrinking
double profileLERP()
{
    extern double Progress, switch_time;
    return std::max(0.,std::min((Progress-switch_time) / (tNorm-switch_time),1.0));
}

double wrap(double x) 
{
	if (x > 1.0)
		x -=1.0;
	if (x < 0.0)
		x += 1.0;
	return x;
}

#define deltaA -0.011
#define deltaB -0.015
#define deltaC -0.005
#define deltaD -0.066

V3d profileXYZ(double ns)
{
    extern double curveLen, oldCurveLen;
    extern double newCurveScale;
    extern double Progress;

    if (curveLen < oldCurveLen) {

        double t1 = profileLERP();
        double t2 = t1*t1;
        double t3 = t2*t1;

        double d1 = 1.-t1;
        double d2 = d1*d1; //(1.-t)^2
        double d3 = d2*d1; //(1.-t)^3

        double lA = wrap(ns+deltaA);
        double lB = wrap(ns+deltaB);
        double lC = wrap(ns+deltaC);
        double lD = wrap(ns+deltaD);


        return newCurveScale * (
                1. * d3 * (V3d) curveXYZ(A,lA)
                +
                3. * t1 * d2 * (V3d) curveXYZ(B,lB)
                +
                3. * t2 * d1 * (V3d) curveXYZ(C,lC)
                +
                1. * t3 * (V3d) curveXYZ(D,lD)
                );

    } else {

        double lA = wrap(ns+deltaA);
        double t = Progress / (0.52*tNorm);
        return newCurveScale * ( (1.-t) * (V3d) curveXYZ(InitContour,lA) + (t) * (V3d) curveXYZ(A,lA));

    }
}

V3d profileNormal(double ns) 
{
    extern double Progress;
    extern double curveLen, oldCurveLen;

    if (curveLen < oldCurveLen) {

        double t1 = profileLERP();
        double t2 = t1*t1;
        double t3 = t2*t1;
        double t4 = t3*t2;

        double d1 = 1.-t1;
        double d2 = d1*d1; //(1.-t)^2
        double d3 = d2*d1; //(1.-t)^3
        double d4 = d3*d2; //(1.-t)^4

        double lA = wrap(ns+deltaA);
        double lB = wrap(ns+deltaB);
        double lC = wrap(ns+deltaC);
        double lD = wrap(ns+deltaD);

        return (
                1. * d3 * (V3d) curveNormal(A,lA)
                +
                3. * t1 * d2 * (V3d) curveNormal(B,lB)
                +
                3. * t2 * d1 * (V3d) curveNormal(C,lC)
                +
                1. * t3 * (V3d) curveNormal(D,lD)
                ).Normalize();
    } else {

        double lA = wrap(ns+deltaA);
        double t = Progress / (0.52*tNorm);
        return ( (1.-t) * (V3d) curveNormal(InitContour,lA) + (t) * (V3d) curveNormal(A,lA));

    }

}

// Compute the length of the current (interpolated) generating curve
double CurveLength()
{
	const double nds = 0.00001; // normalized arc length used for integrating the generating curve length
	double AccumulatedLength = 0.;
	double ns = 0.;
	for (ns=0; ns<1.0-nds; ns+=nds)
  		AccumulatedLength += (profileXYZ(ns+nds) - profileXYZ(ns)).Length();
	AccumulatedLength += (profileXYZ(1.) - profileXYZ(ns)).Length();
	return AccumulatedLength;
}

void curve_SyncCoordinates(primordium *p)
{
    extern double curveLen, oldCurveLen;
    p->pos = profileXYZ(p->s / curveLen);
    p->normal = (profileNormal(p->s / curveLen));
    p->pos += p->n * p->normal;
	return;
}

double curve_PrimordiaDistance (const primordium& P1, const primordium& P2, double C_arclen, double C_radius, bool preciseArcLength=false)
{
    extern double curveLen;
    extern int curve_n;

#if EUCLIDEAN_PHASE_3==1
  extern double Progress, switch_time;
  if (switch_time < Progress) {
    double EuclideanDist = (P1.pos - P2.pos).Length();
    return EuclideanDist;
  } else 
#endif
  {
    double arcLen;
    if (curve_n > 1 && preciseArcLength) {
        const double nds = 0.05; // normalized arc length used for integrating the generating curve length
        double AccumulatedLength = 0.;
        double ns = 0.;
        if (P1.s < P2.s) {
            for (ns=P1.s/curveLen; ns<P2.s/curveLen-nds; ns+=nds)
                AccumulatedLength += (profileXYZ(ns+nds) - profileXYZ(ns)).Length();
            AccumulatedLength += (profileXYZ(P2.s/curveLen) - profileXYZ(ns)).Length();
        } else {
            for (ns=P2.s/curveLen; ns<P1.s/curveLen-nds; ns+=nds) 
                AccumulatedLength += (profileXYZ(ns+nds) - profileXYZ(ns)).Length();
            AccumulatedLength += (profileXYZ(P1.s/curveLen) - profileXYZ(ns)).Length();
        }
        arcLen = std::min(AccumulatedLength,curveLen-AccumulatedLength);
    } else {
        arcLen = std::fabs(P1.s - P2.s); 
        arcLen = std::min(arcLen,curveLen-arcLen);
    }

    // measure distance as: sqrt( arclenth^2 + normal_offset^2)
    double dn = (P1.n * P1.normal - P2.n * P2.normal).Length();

    double dist = std::sqrt(C_arclen*arcLen*arcLen + C_radius*dn*dn);

    return dist;
  }
}

// Check if test primordium tp collides with previous primordia 

int curve_HofmeisterTest(primordium *curve_tp)
{
    extern double curveLen;
    extern double thresholdDistance;
	extern int curve_n;
	double min_dist1 = 1e11;
	double min_dist2 = 1e11+1.0;
	double min_dist3 = 1e11+2.0;
	int min_i1 = curve_n+1;
	int min_i2 = curve_n+2;
	int min_i3 = curve_n+3;

	for (int i = std::max(0,curve_n-MAX_PRIMORDIA_TEST); i<curve_n; i++) {	// iterate over last 100 primordia
			double dist = curve_PrimordiaDistance(*curve_tp,curve_P[i],1.,1.,false);
            if (switch_time < Progress) { 
                if (dist < curve_tp->rho + curve_P[i].rho) 
                    return 1;
                // check again with more precise arc length measurement
                dist = curve_PrimordiaDistance(*curve_tp,curve_P[i],1.,1.,true);
                if (dist < curve_tp->rho + curve_P[i].rho) 
                    return 1;
            } else {
                if (dist < thresholdDistance)
				    return 1;
            }

			// check if this is the closest primordium
			if (dist < min_dist1) {
				min_dist3 = min_dist2;
				min_i3 = min_i2;
				min_dist2 = min_dist1;
				min_i2 = min_i1;
				min_dist1 = dist;
				min_i1 = i;
			} // else check if this is the second closest
			else if (dist < min_dist2) {
				min_dist3 = min_dist2;
				min_i3 = min_i2;
				min_dist2 = dist;
				min_i2 = i;
			}
			else if (dist < min_dist3) {
				min_dist3 = dist;
				min_i3 = i;
			}
	}

    // if there is one primordium or none, the min indicies should all be zero
    // this simplifies assignment of neighbours below
    if (curve_n < 2) {
        min_i1 = min_i2 = min_i3 = 0;
    } 
    else {
        // if the three distances are almost the same,
        // the incipient primordium is tangential to three existing primordia
		if (std::fabs(min_dist2 - min_dist3) < 1e-5 &&
			std::fabs(min_dist1 - min_dist2) < 1e-5 && 
			std::fabs(min_dist1 - min_dist3) < 1e-5) {

            // pick the two youngest
			if (curve_P[min_i1].age > curve_P[min_i2].age && curve_P[min_i1].age > curve_P[min_i3].age) { // min_i1 is oldest
				min_i1 = min_i3;
			}
			else if (curve_P[min_i2].age > curve_P[min_i1].age && curve_P[min_i2].age > curve_P[min_i3].age) { // min_i2 is oldest
				min_i2 = min_i3;
			}
			//else { // min_i3 is oldest
			//}
		}
	}


    // save index of left and right neighbours (anti-clockwise and clockwise, respectively)
    // based on the arc length between them
    double arcLen = curve_P[min_i1].s - curve_tp->s;
    // min1 is CCW neighbour
    if (arcLen >= 0) {
        curve_tp->neighbour_left = min_i1;
        curve_P[min_i1].neighbour_right = curve_n;

        curve_tp->neighbour_right = min_i2;
        curve_P[min_i2].neighbour_left = curve_n;

    } else { // else min2 is CCW neighbour
        curve_tp->neighbour_left = min_i2;
        curve_P[min_i2].neighbour_right = curve_n;

        curve_tp->neighbour_right = min_i1;
        curve_P[min_i1].neighbour_left = curve_n;
    }

	return 0;	// if all primordia were far enough, collision has not occurred
}


module DrawCurve(); 

Axiom: DrawCurve();

// Initializations

double curveLen, oldCurveLen;
double newCurveScale, oldCurveScale; 

Start:
{
	Progress = 0.0;	
    switch_time = 0.;

	thresholdDistance = initialThresholdDistance * USER_FUNC_TH_DIST(Progress);

    // scale curve so its arc length is one
    newCurveScale = 1.;
    curveLen = CurveLength();
    newCurveScale = 1. / curveLen;
    curveLen = CurveLength();
    // set rate of expansion
    newCurveScale = USER_FUNC_AR_RADIUS(0.);
    // save old length and scale
    oldCurveLen = curveLen;
    oldCurveScale = newCurveScale;

    curve_n = 0; // set to zero primordia

    // set first position of trial primordium
    curve_tp.n = 0; // start on the generating curve
	curve_tp.rho = 0.;
	curve_tp.age = 0.;
	curve_tp.neighbour_left = -1; 
	curve_tp.neighbour_right = -1; 
	curve_tp.inserted_dist = thresholdDistance;

    // temp reset receptacle radius
    curve_P[0].n = curve_tp.n;
}

StartEach: {
    // save time when curve starts shrinking inwards.
    if (curveLen >= oldCurveLen) { 
        switch_time = Progress;
    }

	// Advance position of primordia
	for (int i=0; i<curve_n; i++) {
		// advance primordia age
		curve_P[i].age += DAGE * dProgress;

        if (curveLen < oldCurveLen) { 

            // active ring is shrinking but we assume the recepetacle is not growing
            // so we don't change the position of the primordia.
            // but we need to find the arc length coordinate on the active ring

            V3d find_pt = curve_P[i].pos;
            V3d pt1 = profileXYZ(curve_P[i].s / curveLen);
            double s2 = (curve_P[i].s+0.00001);
            if (s2 > curveLen)
                s2 = curveLen - s2;
            V3d pt2 = profileXYZ(s2 / curveLen);
            double dist1 = (pt1-find_pt).Length();
            double dist2 = (pt2-find_pt).Length();
            double ds = curveLen * 0.0001;
            if (dist2 > dist1)
                ds *= -1.;

            int iter = 0;
            while (dist1 > 0.001 && iter < 1e4) {

                curve_P[i].s += ds;
                if (curve_P[i].s > curveLen)
                    curve_P[i].s = curveLen - curve_P[i].s;
                if (curve_P[i].s < 0.)
                    curve_P[i].s = curve_P[i].s + curveLen;

                pt1 = profileXYZ(curve_P[i].s / curveLen);

                dist2 = dist1;
                dist1 = (pt1-find_pt).Length();

                if (dist2 < dist1)
                    ds *= -0.1;

                ++iter;
            }

            // recalculate normal and offset from the found point on the active ring
            V3d pos = profileXYZ(curve_P[i].s / curveLen);
            curve_P[i].normal = (curve_P[i].pos-pos).Normalize();
            curve_P[i].n = (curve_P[i].pos-pos).Length();
        }
        else {
            // if active ring and receptacle are growing,
            // advance primordia centers
            curve_P[i].s *= curveLen/oldCurveLen;
            curve_P[i].n += OFFSET_RATE * dProgress;
            curve_SyncCoordinates(&curve_P[i]); // update normals before lateral displacement
        }


		// recompute radius
		if (curve_n < NUM_BRACTS) {
			curve_P[i].inserted_dist = thresholdDistance;
        }
		curve_P[i].rho = newCurveScale/oldCurveScale * std::sin(curve_P[i].inserted_dist*0.5/(newCurveScale/oldCurveScale));
    }

    if (curveLen >= oldCurveLen) { 
        // Apply lateral displacement (right is clockwise, left is counter clockwise)
        for (int i = 0; i < curve_n; i++) {

            int left_i = curve_P[i].neighbour_left;
            int right_i = curve_P[i].neighbour_right;

            // only incipient primordium can move laterally
            // identified by being younger than both of its neighbours
            if (curve_n > 1 && curve_P[i].age <= curve_P[left_i].age && curve_P[i].age <= curve_P[right_i].age) {

                double dS = 0.;

                if (curve_P[left_i].age >= curve_P[right_i].age) { // CCW is older

                    double d = curve_PrimordiaDistance(curve_P[i],curve_P[left_i],1.,1.,false);
                    dS = k1 * (d - thresholdDistance);

                } else { // CW is older

                    double d = curve_PrimordiaDistance(curve_P[i],curve_P[right_i],1.,1.,false);
                    dS = - k1 * (d - thresholdDistance);

                }

                // update the arc-length coordinate
                double newS = curve_P[i].s + dS * dProgress;
                if (newS < 0.)
                    newS += curveLen;
                if (newS > curveLen)
                    newS -= curveLen;
                curve_P[i].s = newS;

                curve_SyncCoordinates(&curve_P[i]);
            }
        }
    }


	double curve_age = 1.; 
    if (curve_n == 0) // in Model 1, the age of first primordium is 100
      curve_age = 100.;

	// Insert new primordium if they do not collide with previous ones
    for (double s = 0; s < curveLen; s += DS) {
		curve_tp.s = s;
		if (curve_tp.s > curveLen)
			curve_tp.s -= curveLen;

        curve_tp.n = 0;

        // set primordium radius if testing insertion as sum of radii
        curve_tp.rho = newCurveScale/oldCurveScale * std::sin(thresholdDistance*0.5/(newCurveScale/oldCurveScale));
        // set pos if using Euclidean test
        curve_SyncCoordinates(&curve_tp);


		if (curve_HofmeisterTest(&curve_tp) == 0) {
			Printf("Producing primordium %d\n", curve_n+1);

			curve_tp.age = curve_age;
			curve_age += dProgress * DAGE;

			curve_tp.inserted_dist = thresholdDistance;

			curve_P[curve_n] = curve_tp;
    		curve_SyncCoordinates(&curve_P[curve_n]);

			++curve_n;
		}
	}

    // Update size of active ring curve
    if (switch_time >= Progress) {
        oldCurveLen = curveLen;
        oldCurveScale = newCurveScale;
        newCurveScale = USER_FUNC_AR_RADIUS(Progress);
        curveLen = CurveLength();
    }

	// stop simulation when the active ring has advanced to centre of receptacle
    if (profileLERP() > 0.99) {
		Printf ("Simulation done.\n");
		Stop();
	}


	// Reduce the threshold distance for insertion of new primordia
	thresholdDistance = initialThresholdDistance * USER_FUNC_TH_DIST(Progress);

	// Advance time
	Progress += dProgress;

	if (StepNo() % DISPLAY_STEP*int(1./dProgress) == 0) {
		DisplayFrame();
	}
}

derivation length: 1;

interpretation:
maximum depth: 2;

// here the active ring and receptalce are defined by the same curve
// But they should be separate: one curve for active ring and another for the recepetcale.
DrawCurve():
{
    nproduce SetColor(116);

    // draw the curve representing the active ring
    V3d pt = profileXYZ(0.);
    V3d pt2 = profileXYZ(0.001);
    V3f H = (pt2 - pt).Normalize();
    nproduce MoveTo3d(pt) SetWidth(0.1) SetHead3f(H) StartGC;
    for (double u = 0; u < 1.; u += 0.001)
    {
        pt = profileXYZ(u);
        nproduce LineTo3d(pt);
    }
    pt = profileXYZ(1.);
    nproduce LineTo3d(pt);
    nproduce EndGC;

    // draw the primordia
	int color = 0;
    V3d shift(0.,0.,0.1);
    nproduce SB;
	for (int i=0; i<curve_n; i++) {
        color = 8;
        if (i >= NUM_BRACTS) {
            color = 9;
        }
        nproduce MoveTo3d(curve_P[i].pos+shift) SetColor(color) SetWidth(0.05) CircleB(curve_P[i].rho);
    }
    nproduce EB;  
    produce;
}

